#include "VulkanUtils.h"

#include <core/File.h>
#include <core/Logger.h>

static VkInstance instance;
#ifdef DEBUG_VULKAN
static VkDebugUtilsMessengerEXT debugMessenger;
static VkDebugReportCallbackEXT debugReportCallback;
#endif

#define VK_ERROR_CASE(error)  \
    case error: return #error

const char* getVulkanErrorString(VkResult r) {
    switch(r) {
        VK_ERROR_CASE(VK_ERROR_OUT_OF_HOST_MEMORY);
        VK_ERROR_CASE(VK_ERROR_OUT_OF_DEVICE_MEMORY);
        VK_ERROR_CASE(VK_ERROR_INITIALIZATION_FAILED);
        VK_ERROR_CASE(VK_ERROR_LAYER_NOT_PRESENT);
        VK_ERROR_CASE(VK_ERROR_EXTENSION_NOT_PRESENT);
        VK_ERROR_CASE(VK_ERROR_INCOMPATIBLE_DRIVER);
        default: return "unknown";
    }
}

static PFN_vkVoidFunction getVulkanFunction(const char* name) {
    return vkGetInstanceProcAddr(instance, name);
}

#define GET_VULKAN_FUNCTION(name) ((PFN_##name)getVulkanFunction(#name))

bool initVulkanInstance() {
    u32 baseCount = 0;
    const char** baseExtensions = glfwGetRequiredInstanceExtensions(&baseCount);
    if(baseExtensions == nullptr) {
        LOG_ERROR("Could not get required extensions from GLFW");
        return true;
    }
#ifdef DEBUG_VULKAN
    u32 count = baseCount + 2;
    const char* extensions[32];
    if(count > ARRAY_LENGTH(extensions)) {
        LOG_ERROR("Extension buffer is too small");
        return true;
    }
    for(u32 i = 0; i < baseCount; i++) {
        extensions[i] = baseExtensions[i];
    }
    extensions[baseCount] = VK_EXT_DEBUG_UTILS_EXTENSION_NAME;
    extensions[baseCount + 1] = VK_EXT_DEBUG_REPORT_EXTENSION_NAME;
#else
    u32 count = baseCount;
    const char** extensions = baseExtensions;
#endif
    VkApplicationInfo appInfo = {
        .sType = VK_STRUCTURE_TYPE_APPLICATION_INFO,
        .pApplicationName = "Vulkan",
        .applicationVersion = VK_MAKE_VERSION(1, 0, 0),
        .pEngineName = "Kajetan",
        .engineVersion = VK_MAKE_VERSION(0, 0, 1),
        .apiVersion = VK_API_VERSION_1_1};
    VkInstanceCreateInfo info = {
        .sType = VK_STRUCTURE_TYPE_INSTANCE_CREATE_INFO,
        .pApplicationInfo = &appInfo,
#ifdef DEBUG_VULKAN
        .enabledLayerCount = 1,
        .ppEnabledLayerNames = (const char*[]){"VK_LAYER_KHRONOS_validation"},
#endif
        .enabledExtensionCount = count,
        .ppEnabledExtensionNames = extensions};
    VK_ASSERT(vkCreateInstance(&info, nullptr, &instance));
    return false;
}

void destroyVulkanInstance() {
    vkDestroyInstance(instance, nullptr);
    instance = VK_NULL_HANDLE;
}

#ifdef DEBUG_VULKAN
static VKAPI_ATTR VkBool32 onVulkanDebugMessenger VKAPI_CALL(
    VkDebugUtilsMessageSeverityFlagBitsEXT, VkDebugUtilsMessageTypeFlagsEXT,
    const VkDebugUtilsMessengerCallbackDataEXT* data, void*) {
    LOG_WARNING("Vulkan validation layer message: %s", data->pMessage);
    return false;
}

static bool initDebugMessenger() {
    VkDebugUtilsMessengerCreateInfoEXT info = {
        .sType = VK_STRUCTURE_TYPE_DEBUG_UTILS_MESSENGER_CREATE_INFO_EXT,
        .messageSeverity = VK_DEBUG_UTILS_MESSAGE_SEVERITY_WARNING_BIT_EXT |
                           VK_DEBUG_UTILS_MESSAGE_SEVERITY_ERROR_BIT_EXT,
        .messageType = VK_DEBUG_UTILS_MESSAGE_TYPE_GENERAL_BIT_EXT |
                       VK_DEBUG_UTILS_MESSAGE_TYPE_VALIDATION_BIT_EXT |
                       VK_DEBUG_UTILS_MESSAGE_TYPE_PERFORMANCE_BIT_EXT,
        .pfnUserCallback = onVulkanDebugMessenger};
    auto f = GET_VULKAN_FUNCTION(vkCreateDebugUtilsMessengerEXT);
    if(f == nullptr) {
        LOG_WARNING("Could not find debug util messenger function");
        return false;
    }
    VK_ASSERT(f(instance, &info, nullptr, &debugMessenger));
    return false;
}

static VKAPI_ATTR VkBool32 onVulkanDebugReport VKAPI_CALL(
    VkDebugReportFlagsEXT, VkDebugReportObjectTypeEXT, uint64_t, size_t,
    int32_t, const char* pLayerPrefix, const char* pMessage, void*) {
    LOG_WARNING("Vulkan debug message '%s': %s", pLayerPrefix, pMessage);
    return false;
}

static bool initDebugReportCallback() {
    VkDebugReportCallbackCreateInfoEXT info = {
        .sType = VK_STRUCTURE_TYPE_DEBUG_REPORT_CREATE_INFO_EXT,
        .flags = VK_DEBUG_REPORT_WARNING_BIT_EXT |
                 VK_DEBUG_REPORT_PERFORMANCE_WARNING_BIT_EXT |
                 VK_DEBUG_REPORT_ERROR_BIT_EXT,
        .pfnCallback = onVulkanDebugReport};
    auto f = GET_VULKAN_FUNCTION(vkCreateDebugReportCallbackEXT);
    if(f == nullptr) {
        LOG_WARNING("Could not find debug report function");
        return false;
    }
    VK_ASSERT(f(instance, &info, nullptr, &debugReportCallback));
    return false;
}

bool initVulkanDebugging() {
    return initDebugMessenger() || initDebugReportCallback();
}

static void destroyDebugMessenger() {
    auto f = GET_VULKAN_FUNCTION(vkDestroyDebugUtilsMessengerEXT);
    if(f == nullptr) {
        LOG_WARNING("Could not find debug util messenger destroy function");
        return;
    }
    f(instance, debugMessenger, nullptr);
}

static void destroyDebugReportCallback() {
    auto f = GET_VULKAN_FUNCTION(vkDestroyDebugReportCallbackEXT);
    if(f == nullptr) {
        LOG_WARNING("Could not find debug report destroy function");
        return;
    }
    f(instance, debugReportCallback, nullptr);
}

void destroyVulkanDebugging() {
    if(instance != VK_NULL_HANDLE) {
        destroyDebugMessenger();
        destroyDebugReportCallback();
    }
}
#else
bool initVulkanDebugging() {
    (void)getVulkanFunction;
    return false;
}

void destroyVulkanDebugging() {
}
#endif

u32 findVulkanQueueFamily(VkPhysicalDevice pd, VkQueueFlags flags) {
    VkQueueFamilyProperties properties[32];
    u32 count = ARRAY_LENGTH(properties);
    vkGetPhysicalDeviceQueueFamilyProperties(pd, &count, properties);
    for(u32 i = 0; i < count; i++) {
        if((properties[i].queueFlags & flags) == flags) {
            return i;
        }
    }
    return INVALID_VULKAN_QUEUE_FAMILY;
}

static bool noPresentationSupport(
    VkPhysicalDevice pd, VkSurfaceKHR s, u32 index) {
    VkBool32 b = false;
    VK_ASSERT(vkGetPhysicalDeviceSurfaceSupportKHR(pd, index, s, &b));
    return !b;
}

u32 findVulkanSurfaceQueueFamily(VkPhysicalDevice pd, VkSurfaceKHR s) {
    u32 count = 0;
    vkGetPhysicalDeviceQueueFamilyProperties(pd, &count, nullptr);
    for(u32 i = 0; i < count; i++) {
        if(!noPresentationSupport(pd, s, i)) {
            return i;
        }
    }
    return INVALID_VULKAN_QUEUE_FAMILY;
}

bool findVulkanSurfaceFormat(
    VkSurfaceFormatKHR* sf, VkPhysicalDevice pd, VkSurfaceKHR s,
    VulkanSurfaceFormatSelector sfs) {
    VkSurfaceFormatKHR formats[64];
    u32 c = ARRAY_LENGTH(formats);
    VK_ASSERT(vkGetPhysicalDeviceSurfaceFormatsKHR(pd, s, &c, formats));
    int bestPoints = 0;
    for(u32 i = 0; i < c; i++) {
        int points = sfs(formats + i);
        if(points > bestPoints) {
            bestPoints = points;
            *sf = formats[i];
        }
    }
    return bestPoints == 0;
}

bool findVulkanSurfacePresentMode(
    VkPresentModeKHR* m, VkPhysicalDevice pd, VkSurfaceKHR s,
    VulkanSurfacePresentModeSelector spms) {
    VkPresentModeKHR modes[64];
    u32 c = ARRAY_LENGTH(modes);
    VK_ASSERT(vkGetPhysicalDeviceSurfacePresentModesKHR(pd, s, &c, modes));
    int bestPoints = 0;
    for(u32 i = 0; i < c; i++) {
        int points = spms(modes[i]);
        if(points > bestPoints) {
            bestPoints = points;
            *m = modes[i];
        }
    }
    return bestPoints == 0;
}

bool findVulkanPhysicalDevice(
    VkPhysicalDevice* pd, VulkanPhysicalDeviceSelector s) {
    VkPhysicalDevice devices[32];
    u32 c = ARRAY_LENGTH(devices);
    VK_ASSERT(vkEnumeratePhysicalDevices(instance, &c, devices));
    int bestPoints = 0;
    for(u32 i = 0; i < c; i++) {
        int points = s(devices[i]);
        if(points > bestPoints) {
            bestPoints = points;
            *pd = devices[i];
        }
    }
    return bestPoints == 0;
}

bool hasVulkanExtension(VkPhysicalDevice pd, const char* extension) {
    VkExtensionProperties e[1024];
    u32 c = ARRAY_LENGTH(e);
    VkResult r = vkEnumerateDeviceExtensionProperties(pd, nullptr, &c, e);
    if(r != VK_SUCCESS) {
        LOG_ERROR(
            "Cannot get physical device extension properties: %s",
            getVulkanErrorString(r));
        return false;
    }
    for(u32 i = 0; i < c; i++) {
        if(strcmp(e[i].extensionName, extension) == 0) {
            return true;
        }
    }
    return false;
}

bool initVulkanDevice(
    VkDevice* d, VkPhysicalDevice pd, const VulkanDeviceQueueData* data,
    size_t n, const char** extensions, size_t nExtensions) {
    constexpr size_t LENGTH = 32;
    if(n >= LENGTH) {
        LOG_ERROR("Vulkan device queue overload");
        return true;
    }
    VkDeviceQueueCreateInfo deviceQueueInfo[LENGTH] = {0};
    for(size_t i = 0; i < n; i++) {
        VkDeviceQueueCreateInfo* qInfo = deviceQueueInfo + i;
        qInfo->sType = VK_STRUCTURE_TYPE_DEVICE_QUEUE_CREATE_INFO;
        qInfo->queueFamilyIndex = data[i].queueFamilyIndex;
        qInfo->queueCount = 1;
        qInfo->pQueuePriorities = &data[i].priority;
    }
    VkPhysicalDeviceFeatures deviceFeatures = {0};
    vkGetPhysicalDeviceFeatures(pd, &deviceFeatures);
    VkDeviceCreateInfo info = {
        .sType = VK_STRUCTURE_TYPE_DEVICE_CREATE_INFO,
        .queueCreateInfoCount = (u32)n,
        .pQueueCreateInfos = deviceQueueInfo,
        .enabledExtensionCount = (u32)nExtensions,
        .ppEnabledExtensionNames = extensions,
        .pEnabledFeatures = &deviceFeatures};
    VK_ASSERT(vkCreateDevice(pd, &info, nullptr, d));
    return false;
}

void destroyVulkanDevice(VkDevice d) {
    if(instance != VK_NULL_HANDLE) {
        vkDestroyDevice(d, nullptr);
    }
}

bool initVulkanSurface(VkSurfaceKHR* s, GLFWwindow* w) {
    VK_ASSERT(glfwCreateWindowSurface(instance, w, nullptr, s));
    return false;
}

void destroyVulkanSurface(VkSurfaceKHR s) {
    if(instance != VK_NULL_HANDLE) {
        vkDestroySurfaceKHR(instance, s, nullptr);
    }
}

static u32 getSwapImageCount(const VkSurfaceCapabilitiesKHR* caps) {
    u32 c = caps->minImageCount + 1;
    // according to VkSurfaceCapabilitiesKHR doc:
    // maxImageCount is 0 when there is no strict limit
    if(caps->maxImageCount != 0) {
        return minU32(c, caps->maxImageCount);
    }
    return c;
}

bool initVulkanSwapchain(VkSwapchainKHR* sc, VulkanSwapchainData* d) {
    VkSurfaceCapabilitiesKHR caps = {0};
    VK_ASSERT(vkGetPhysicalDeviceSurfaceCapabilitiesKHR(
        d->physicalDevice, d->surface, &caps));
    VkSwapchainCreateInfoKHR ci = {
        .sType = VK_STRUCTURE_TYPE_SWAPCHAIN_CREATE_INFO_KHR,
        .surface = d->surface,
        .minImageCount = getSwapImageCount(&caps),
        .imageFormat = d->surfaceFormat.format,
        .imageColorSpace = d->surfaceFormat.colorSpace,
        .imageExtent = d->size,
        .imageArrayLayers = 1,
        .imageUsage = VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT,
        .imageSharingMode = d->sharingMode,
        .queueFamilyIndexCount = d->queueFamilyIndexCount,
        .pQueueFamilyIndices = d->queueFamilyIndices,
        .preTransform = caps.currentTransform,
        .compositeAlpha = VK_COMPOSITE_ALPHA_OPAQUE_BIT_KHR,
        .presentMode = d->presentMode,
        .clipped = VK_TRUE,
        .oldSwapchain = VK_NULL_HANDLE};
    VK_ASSERT(vkCreateSwapchainKHR(d->device, &ci, nullptr, sc));
    return false;
}

void destroyVulkanSwapchain(VkSwapchainKHR s, VkDevice d) {
    if(d != VK_NULL_HANDLE) {
        vkDestroySwapchainKHR(d, s, nullptr);
    }
}

static bool createImageView(
    VkImageView* view, VkDevice d, VkImage image, VkFormat format) {
    VkImageViewCreateInfo info = {
        .sType = VK_STRUCTURE_TYPE_IMAGE_VIEW_CREATE_INFO,
        .image = image,
        .viewType = VK_IMAGE_VIEW_TYPE_2D,
        .format = format,
        .subresourceRange = {
            .aspectMask = VK_IMAGE_ASPECT_COLOR_BIT,
            .baseMipLevel = 0,
            .levelCount = 1,
            .baseArrayLayer = 0,
            .layerCount = 1}};
    VK_ASSERT(vkCreateImageView(d, &info, nullptr, view));
    return false;
}

bool initVulkanSwapchainImages(
    VulkanSwapchainImages* si, VkDevice d, VkSwapchainKHR sc, VkFormat format) {
    VK_ASSERT(vkGetSwapchainImagesKHR(d, sc, &si->amount, nullptr));
    si->images = coreZeroAllocate(sizeof(VkImage) * si->amount);
    si->imageViews = coreZeroAllocate(sizeof(VkImageView) * si->amount);
    VK_ASSERT(vkGetSwapchainImagesKHR(d, sc, &si->amount, si->images));
    for(u32 x = 0; x < si->amount; x++) {
        if(createImageView(si->imageViews + x, d, si->images[x], format)) {
            return true;
        }
    }
    return false;
}

void destroyVulkanSwapchainImages(VulkanSwapchainImages* si, VkDevice d) {
    if(d == VK_NULL_HANDLE || si == nullptr) {
        return;
    }
    for(size_t x = 0; x < si->amount; x++) {
        vkDestroyImageView(d, si->imageViews[x], nullptr);
    }
    coreFree(si->images);
    coreFree(si->imageViews);
    *si = (VulkanSwapchainImages){0};
}

bool initVulkanSemaphore(VkSemaphore* s, VkDevice d) {
    VkSemaphoreCreateInfo info = {
        .sType = VK_STRUCTURE_TYPE_SEMAPHORE_CREATE_INFO};
    VK_ASSERT(vkCreateSemaphore(d, &info, nullptr, s));
    return false;
}

void destroyVulkanSemaphore(VkSemaphore s, VkDevice d) {
    if(d != VK_NULL_HANDLE) {
        vkDestroySemaphore(d, s, nullptr);
    }
}

static bool initShaderModule(FileContent* f, VkDevice d, VkShaderModule* sm) {
    if((f->length % 4) != 0) {
        LOG_ERROR("Shader size is not a multiple of 4");
        return true;
    }
    VkShaderModuleCreateInfo info = {
        .sType = VK_STRUCTURE_TYPE_SHADER_MODULE_CREATE_INFO,
        .codeSize = f->length,
        .pCode = (u32*)f->data};
    VK_ASSERT(vkCreateShaderModule(d, &info, nullptr, sm));
    return false;
}

bool initVulkanShaderModule(VkShaderModule* sm, VkDevice d, const char* path) {
    FileContent f = {0};
    if(readFile(&f, path)) {
        return true;
    }
    bool r = initShaderModule(&f, d, sm);
    destroyFileContent(&f);
    return r;
}

void destroyVulkanShaderModule(VkShaderModule sm, VkDevice d) {
    if(d != VK_NULL_HANDLE) {
        vkDestroyShaderModule(d, sm, nullptr);
    }
}

bool initVulkanPipelineLayout(VkPipelineLayout* pl, VkDevice d) {
    VkPipelineLayoutCreateInfo info = {
        .sType = VK_STRUCTURE_TYPE_PIPELINE_LAYOUT_CREATE_INFO,
        .setLayoutCount = 0,
        .pSetLayouts = nullptr,
        .pushConstantRangeCount = 0,
        .pPushConstantRanges = nullptr};
    VK_ASSERT(vkCreatePipelineLayout(d, &info, nullptr, pl));
    return false;
}

void destroyVulkanPipelineLayout(VkPipelineLayout pl, VkDevice d) {
    if(d != VK_NULL_HANDLE) {
        vkDestroyPipelineLayout(d, pl, nullptr);
    }
}

bool initVulkanFramebuffers(
    VkFramebuffer** f, VulkanSwapchainImages* si, VkDevice d, VkRenderPass rp,
    u32 width, u32 height) {
    *f = coreZeroAllocate(sizeof(VkFramebuffer) * si->amount);
    for(u32 i = 0; i < si->amount; i++) {
        VkFramebufferCreateInfo info = {
            .sType = VK_STRUCTURE_TYPE_FRAMEBUFFER_CREATE_INFO,
            .renderPass = rp,
            .attachmentCount = 1,
            .pAttachments = si->imageViews + i,
            .width = width,
            .height = height,
            .layers = 1};
        VK_ASSERT(vkCreateFramebuffer(d, &info, nullptr, (*f) + i));
    }
    return false;
}

void destroyVulkanFramebuffers(VkFramebuffer** f, u32 amount, VkDevice d) {
    if(f != nullptr) {
        if(d != VK_NULL_HANDLE && *f != nullptr) {
            for(u32 i = 0; i < amount; i++) {
                vkDestroyFramebuffer(d, (*f)[i], nullptr);
            }
        }
        coreFree(*f);
        *f = nullptr;
    }
}

/*static bool initCommandBuffers() {
    vk.swapchain.commandsBuffers =
        coreAllocate(sizeof(VkCommandBuffer) * vk.swapchain.amount);
    VkCommandBufferAllocateInfo a = {
        .sType = VK_STRUCTURE_TYPE_COMMAND_BUFFER_ALLOCATE_INFO,
        .commandPool = vk.swapchain.commandPool,
        .level = VK_COMMAND_BUFFER_LEVEL_PRIMARY,
        .commandBufferCount = vk.swapchain.amount};
    VK_ASSERT(vkAllocateCommandBuffers(
        vk.device.logical, &a, vk.swapchain.commandsBuffers));
    return false;
}

static void destroyCommandBuffers() {
    if(vk.swapchain.commandPool != VK_NULL_HANDLE && vk.swapchain.amount >
0) { vkFreeCommandBuffers( vk.device.logical, vk.swapchain.commandPool,
vk.swapchain.amount, vk.swapchain.commandsBuffers);
    }
    coreFree(vk.swapchain.commandsBuffers);
}

static bool fillCommandBuffer(u32 index) {
    (void)index;
    // VkCommandBufferBeginInfo info = {
    //     .sType = VK_STRUCTURE_TYPE_COMMAND_BUFFER_BEGIN_INFO,
    //     .flags = VK_COMMAND_BUFFER_USAGE_SIMULTANEOUS_USE_BIT};
    // VK_ASSERT(vkBeginCommandBuffer(vk.swapchain.commandsBuffers[index],
    // &info)); VkClearValue v = {.color = {.int32 = {255, 0, 0, 0}}};
    // VkRenderPassBeginInfo rInfo = {
    //     .sType = VK_STRUCTURE_TYPE_RENDER_PASS_BEGIN_INFO,
    //     .renderPass = vk.renderPass,
    //     .framebuffer = vk.swapchain.framebuffers[index],
    //     .renderArea = {.offset = {0, 0}, .extent = {vk.width,
vk.height}},
    //     .clearValueCount = 1,
    //     .pClearValues = &v};
    // vkCmdBeginRenderPass(vk.swapchain.commandsBuffers[index], &rInfo,
    //                      VK_SUBPASS_CONTENTS_INLINE);
    return false;
}*/
